#!/usr/bin/env python3
# ----------------------------------------------------------------------------
#
# Copyright 2018 EMVA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------


# Standard library imports
import io
import datetime
import pathlib
from threading import Lock
import time
import zipfile

# Related third party imports
import numpy as np

from genicam2.genapi import NodeMap
from genicam2.genapi import LogicalErrorException
from genicam2.gentl import TimeoutException, AccessDeniedException, \
    LoadLibraryException, InvalidParameterException, \
    NotImplementedException, NotAvailableException, ClosedException, \
    ResourceInUseException
from genicam2.gentl import GenTLProducer, BufferToken, EventManagerNewBuffer
from genicam2.gentl import DEVICE_ACCESS_FLAGS_LIST, EVENT_TYPE_LIST, \
    ACQ_START_FLAGS_LIST, ACQ_STOP_FLAGS_LIST, ACQ_QUEUE_TYPE_LIST, \
    TL_CHAR_ENCODING_LIST

# Local application/library specific imports
from harvesters._private.core.buffer import Buffer
from harvesters._private.core.port import ConcretePort
from harvesters._private.core.statistics import Statistics
from harvesters._private.core.thread import PyThread
from harvesters._private.core.thread_ import MutexLocker
from harvesters.processor import Processor
from harvesters._private.core.pfnc import symbolics
from harvesters._private.core.pfnc import uint8_formats, uint16_formats


class _ProcessorConvertPyBytesToNumpy1D(Processor):
    def __init__(self):
        #
        super().__init__(
            description='Converts a Python bytes object to a Numpy 1D array'
        )

    def process(self, input_buffer: Buffer):
        symbolic = None
        try:
            pixel_format_int = input_buffer.gentl_buffer.pixel_format
        except InvalidParameterException:
            pass
        else:
            symbolic = symbolics[int(pixel_format_int)]

        if symbolic in uint8_formats:
            dtype = 'uint8'
        elif symbolic in uint16_formats:
            dtype = 'uint16'
        else:
            dtype = 'uint8'

        output_buffer = Buffer(
            data_stream=input_buffer.data_stream,
            gentl_buffer=input_buffer.gentl_buffer,
            node_map=input_buffer.node_map,
            image=np.frombuffer(
                input_buffer.gentl_buffer.raw_buffer, dtype=dtype
            )
        )
        return output_buffer


class Harvester:
    def __init__(self, frontend=None, profile=False, min_num_buffers=16, parent=None):
        """
        Is a Python class that works as Harvester Core. You can image
        acquisition related task through this class.

        :param frontend:
        :param profile:
        :param min_num_buffers:
        :param parent:
        """
        #
        super().__init__()

        #
        self._parent = parent

        #
        self._frontend = frontend

        #
        self._is_acquiring_images = False

        #
        self._cti_files = []
        self._producers = []
        self._systems = []
        self._interfaces = []
        self._device_info_list = []
        self._concrete_port = None

        #
        self._announced_buffers = []
        self._fetched_buffers = []

        self._data_stream = None
        self._event_manager = None

        self._device = None

        #
        self._min_num_buffers = min_num_buffers
        self._num_extra_buffers = 1

        #
        self._has_revised_device_list = False
        self._timeout_for_update = 1000  # ms
        self._has_acquired_1st_image = False

        #
        self._mutex = Lock()
        self._thread_image_acquisition = PyThread(
            mutex=self._mutex,
            worker=self._worker_image_acquisition
        )
        self._thread_statistics_measurement = PyThread(
            mutex=self._mutex,
            worker=self._worker_acquisition_statistics
        )

        #
        self._feature_tree_model = None

        #
        self._current_width = 0
        self._current_height = 0
        self._current_pixel_format = ''

        self._updated_statistics = None
        self._signal_stop_image_acquisition = None

        #
        self._statistics_update_cycle = 1  # s
        self._statistics_latest = Statistics()
        self._statistics_overall = Statistics()
        self._statistics_list = [
            self._statistics_latest, self._statistics_overall
        ]

        #
        self._timeout_for_image_acquisition = 1  # ms

        #
        self._processors = []
        self._system_defined_processors = [_ProcessorConvertPyBytesToNumpy1D()]
        self._user_defined_processors = []

        #
        self._num_images_to_acquire = -1
        self._commands = []

        #
        if profile:
            from harvesters._private.core.helper import Profiler
            self._profiler = Profiler()
        else:
            self._profiler = None

        if self._profiler:
            self._profiler.print_diff()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reset()

    @property
    def updated_statistics(self):
        return self._updated_statistics

    @updated_statistics.setter
    def updated_statistics(self, obj):
        self._updated_statistics = obj

    @property
    def signal_stop_image_acquisition(self):
        return self._signal_stop_image_acquisition

    @signal_stop_image_acquisition.setter
    def signal_stop_image_acquisition(self, obj):
        self._signal_stop_image_acquisition = obj

    @property
    def device(self):
        return self._device

    @property
    def cti_files(self):
        return self._cti_files

    @property
    def is_acquiring_images(self):
        return self._is_acquiring_images

    @property
    def device_info_list(self):
        return self._device_info_list

    @property
    def timeout_for_update(self):
        return self._timeout_for_update

    @timeout_for_update.setter
    def timeout_for_update(self, ms):
        self._timeout_for_update = ms

    @property
    def timeout_for_image_acquisition(self):
        return self._timeout_for_image_acquisition

    @timeout_for_image_acquisition.setter
    def timeout_for_image_acquisition(self, ms):
        with MutexLocker(self.thread_image_acquisition):
            self._timeout_for_image_acquisition = ms

    @property
    def user_defined_processors(self):
        return self._user_defined_processors

    @property
    def has_revised_device_info_list(self):
        return self._has_revised_device_list

    @has_revised_device_info_list.setter
    def has_revised_device_info_list(self, value):
        self._has_revised_device_list = value

    @property
    def thread_image_acquisition(self):
        return self._thread_image_acquisition

    @thread_image_acquisition.setter
    def thread_image_acquisition(self, obj):
        self._thread_image_acquisition = obj
        self._thread_image_acquisition.worker = self._worker_image_acquisition

    @property
    def thread_statistics_measurement(self):
        return self._thread_statistics_measurement

    @thread_statistics_measurement.setter
    def thread_statistics_measurement(self, obj):
        self._thread_statistics_measurement = obj
        self._thread_statistics_measurement.worker = self._worker_acquisition_statistics

    def connect_device(self, item_id=0, id_info=None, model=None,
            serial_number=None, user_defined_name=None, vendor=None):
        """
        Connect the specified device to the Harvester object.

        :param item_id:
        :param id_info:
        :param model:
        :param serial_number:
        :param user_defined_name:
        :param vendor:

        :return: None.
        """
        #
        if self.device or self.device_info_list is None:
            # TODO: Throw an exception to tell clients that there's no
            # device to connect.
            return

        # Instantiate a GenTL Device module.
        self._device = \
            device=self.device_info_list[item_id].create_device()

        # Then open it.
        try:
            self.device.open(
                DEVICE_ACCESS_FLAGS_LIST.DEVICE_ACCESS_EXCLUSIVE
            )
        except AccessDeniedException as e:
            print(e)
            self.disconnect_device()
        else:
            # And get an alias of its GenTL Port module.
            port = self.device.remote_port

            # Inquire it's URL information.
            # TODO: Consider a case where len(url_info_list) > 1.
            url = port.url_info_list[0].url

            # And parse the URL.
            location, others = url.split(':', 1)
            file_name, address, size = others.split(';')
            address = int(address, 16)

            # It may specify the schema version.
            delimiter = '?'
            if delimiter in size:
                size, _ = size.split(delimiter)
            size = int(size, 16)

            # Now we get the file content.
            content = port.read(address, size)

            # But wait, we have to check if it's a zip file or not.
            content = content[1]
            file_content = io.BytesIO(content)

            # Let's check the reality.
            if zipfile.is_zipfile(file_content):
                # Yes, that's a zip file.
                file_content = zipfile.ZipFile(file_content, 'r')

                # Extract the file content from the zip file.
                for file_info in file_content.infolist():
                    if pathlib.Path(
                            file_info.filename).suffix.lower() == '.xml':
                        #
                        content = file_content.read(file_info).decode('utf8')
                        break

            # Instantiate a GenICam node map object.
            self.device.node_map = NodeMap()

            # Then load the XML file content on the node map object.
            self.device.node_map.load_xml_from_string(content)

            # Instantiate a concrete port object of the remote device's
            # port.
            self._concrete_port = ConcretePort(self.device.remote_port)

            # And finally connect the concrete port on the node map
            # object.
            self.device.node_map.connect(self._concrete_port, port.name)

            if self._profiler:
                self._profiler.print_diff()

    def disconnect_device(self):
        """
        Disconnect the device that has been connected to the Harvester
        object.

        :return: None.
        """
        #
        if self.device:
            if self.device.is_open():
                self.stop_image_acquisition()
                self.device.close()

            if self.device.node_map:
                self.device.node_map.disconnect()
            self.device.node_map = None

        self._device = None
        self._concrete_port = None

        if self._profiler:
            self._profiler.print_diff()

    def start_image_acquisition(self):
        """
        Starts image acquisition.

        :return: None.
        """
        if self.is_acquiring_images:
            # If it's pausing drawing images, just resume it and
            # immediately return this method.
            if self._frontend:
                if self._frontend.canvas.is_pausing:
                    self._frontend.canvas.resume_drawing()
        else:
            #
            self._data_stream = self.device.create_data_stream()
            self._data_stream.open(self.device.data_stream_ids[0])

            #
            num_required_buffers = self._min_num_buffers
            try:
                num_buffers = self._data_stream.buffer_announce_min
                if num_buffers < num_required_buffers:
                    num_buffers = num_required_buffers
            except InvalidParameterException as e:
                num_buffers = num_required_buffers

            if self._data_stream.defines_payload_size():
                buffer_size = self._data_stream.payload_size
            else:
                buffer_size = self.device.node_map.PayloadSize.value

            raw_buffers = self._create_raw_buffers(
                num_buffers, buffer_size
            )
            buffer_tokens = self._create_buffer_tokens(
                raw_buffers
            )

            self._announced_buffers = self._announce_buffers(
                buffer_tokens
            )
            self._queue_announced_buffers(self._announced_buffers)

            #
            event_token = self._data_stream.register_event(
                EVENT_TYPE_LIST.EVENT_NEW_BUFFER
            )
            self._event_manager = EventManagerNewBuffer(event_token)

            # Reset the number of images to acquire.
            try:
                acq_mode = self.device.node_map.AcquisitionMode.value
                if acq_mode == 'Continuous':
                    num_images_to_acquire = -1
                elif acq_mode == 'SingleFrame':
                    num_images_to_acquire = 1
                elif acq_mode == 'MultiFrame':
                    num_images_to_acquire = self.node_map.AcquisitionFrameCount.value
                else:
                    num_images_to_acquire = -1
            except LogicalErrorException:
                # The node doesn't exist.
                num_images_to_acquire = -1

            self._num_images_to_acquire = num_images_to_acquire

            # Update the sequence of image processors.
            self._processors = []

            for p in self._system_defined_processors:
                self._processors.append(p)

            for p in self.user_defined_processors:
                self._processors.append(p)

            # Start image acquisition.
            self._data_stream.start_acquisition(
                ACQ_START_FLAGS_LIST.ACQ_START_FLAGS_DEFAULT,
                self._num_images_to_acquire
            )

            #
            self._is_acquiring_images = True

            #
            self.reset_statistics()
            if self.thread_statistics_measurement:
                self.thread_statistics_measurement.start()

            #
            if self.thread_image_acquisition:
                self.thread_image_acquisition.start()

            #
            self.device.node_map.AcquisitionStart.execute()

    def _worker_acquisition_statistics(self):
        if not self.is_acquiring_images:
            return

        time.sleep(self._statistics_update_cycle)

        with MutexLocker(self.thread_statistics_measurement):
            #
            if self._frontend:
                #
                message_config = 'W: {0} x H: {1}, {2}, '.format(
                    self._current_width,
                    self._current_height,
                    self._current_pixel_format,
                )

                #
                message_latest = ''
                if self._statistics_latest.num_images > 0:
                    message_latest = '{0:.1f} fps in the last {1:.1f} s, '.format(
                        self._statistics_latest.fps,
                        self._statistics_update_cycle
                    )

                #
                message_overall = '{0:.1f} fps in the last {1}, {2} images'.format(
                    self._statistics_overall.fps,
                    str(datetime.timedelta(
                        seconds=int(self._statistics_overall.elapsed_time_s)
                    )),
                    self._statistics_overall.num_images
                )

                #
                if self.updated_statistics:
                    self.updated_statistics.emit(
                        '{0}'.format(
                            message_config + message_latest + message_overall
                        )
                    )

            self._statistics_latest.reset()

    def _worker_image_acquisition(self):
        try:
            if self.is_acquiring_images:
                time.sleep(0.001)
                self._event_manager.update_event_data(
                    self._timeout_for_image_acquisition
                )
            else:
                return
        except TimeoutException as e:
            pass
        else:
            #
            if not self.is_acquiring_images:
                return

            gentl_buffer = self._event_manager.buffer

            #
            self._update_statistics(gentl_buffer)

            # Put the buffer in the process flow.
            input_buffer = Buffer(
                self._data_stream, gentl_buffer, self.device.node_map, None
            )
            output_buffer = None

            for p in self._processors:
                output_buffer = p.process(input_buffer)
                input_buffer = output_buffer

            if output_buffer:
                # We've got a new image so now we can reuse the buffer that
                # we had kept.
                with MutexLocker(self.thread_image_acquisition):
                    if len(self._fetched_buffers) > 0:
                        # We have a buffer now so we queue it; it's discarded
                        # before being used.
                        self.queue_buffer(self._fetched_buffers.pop(0))

                    if output_buffer.image.ndarray is not None:
                        # Append the recently fetched buffer.
                        # Then one buffer remains for our client.
                        self._fetched_buffers.append(output_buffer)

            #
            if self._num_images_to_acquire >= 1:
                self._num_images_to_acquire -= 1

            if self._num_images_to_acquire == 0:
                #
                if self.signal_stop_image_acquisition:
                    self.signal_stop_image_acquisition.emit()

    def fetch_buffer(self, timeout_ms=0):
        """
        Fetches the latest Buffer object and returns it.

        :param timeout_ms: Set timeout value in ms.
        :return: A Buffer object.
        """
        if not self.is_acquiring_images:
            return None

        watch_timeout = True if timeout_ms > 0 else False
        buffer = None
        base = time.time()

        while buffer is None:
            if watch_timeout and (time.time() - base) > timeout_ms:
                break
            else:
                with MutexLocker(self.thread_image_acquisition):
                    if len(self._fetched_buffers) > 0:
                        buffer = self._fetched_buffers.pop(0)

        return buffer

    def queue_buffer(self, buffer):
        """
        Queues the Buffer object.

        :param buffer: Set a Buffer object to queue.

        :return: None.
        """
        if self._data_stream and buffer:
            self._data_stream.queue_buffer(
                buffer.gentl_buffer
            )

    def _update_statistics(self, gentl_buffer):
        #
        frequency = 1000000000.  # Hz
        try:
            timestamp = gentl_buffer.timestamp_ns
        except (InvalidParameterException, NotImplementedException, NotAvailableException):
            try:
                # The unit is device/implementation dependent.
                timestamp = gentl_buffer.timestamp
            except (NotImplementedException, NotAvailableException):
                timestamp = 0  # Not available
            else:
                try:
                    frequency = self.device.timestamp_frequency
                except (InvalidParameterException, NotImplementedException,
                        NotAvailableException):
                    try:
                        frequency = self.device.node_map.GevTimestampTickFrequency.value
                    except LogicalErrorException:
                        pass

        #
        for statistics in self._statistics_list:
            statistics.increment_num_images()
            statistics.set_timestamp(timestamp, frequency)

    @staticmethod
    def _create_raw_buffers(num_buffers, size):
        # Instantiate a list object.
        raw_buffers = []

        # Append bytes objects to the list.
        # The number is specified by num_buffer and the buffer size is
        # specified by size.
        for _ in range(num_buffers):
            raw_buffers.append(bytes(size))

        # Then return the list.
        return raw_buffers

    @staticmethod
    def _create_buffer_tokens(raw_buffers):
        # Instantiate a list object.
        _buffer_tokens = []

        # Append Buffer Token object to the list.
        for i, buffer in enumerate(raw_buffers):
            _buffer_tokens.append(
                BufferToken(buffer, i)
            )

        # Then returns the list.
        return _buffer_tokens

    def _announce_buffers(self, _buffer_tokens):
        #
        announced_buffers = []

        # Iterate announcing buffers in the Buffer Tokens.
        for token in _buffer_tokens:
            # Get an announced buffer.
            announced_buffer = self._data_stream.announce_buffer(token)

            # And append it to the list.
            announced_buffers.append(announced_buffer)

        # Then return the list of announced Buffer objects.
        return announced_buffers

    def _queue_announced_buffers(self, buffers):
        for buffer in buffers:
            self._data_stream.queue_buffer(buffer)

    def stop_image_acquisition(self):
        """
        Stops image acquisition.

        :return: None.
        """
        if self.is_acquiring_images:
            #
            self._is_acquiring_images = False

            #
            if self.thread_image_acquisition.is_running:
                self.thread_image_acquisition.stop()

            if self.thread_statistics_measurement.is_running:
                self.thread_statistics_measurement.stop()

            with MutexLocker(self.thread_image_acquisition):
                #
                self._event_manager.flush_event_queue()

                # Stop image acquisition.
                try:
                    self._data_stream.stop_acquisition(
                        ACQ_STOP_FLAGS_LIST.ACQ_STOP_FLAGS_KILL
                    )
                except ResourceInUseException:
                    # Device throw RESOURCE_IN_USE exception
                    # if the acquisition has already terminated or
                    # it has not been started.
                    pass
                except TimeoutException as e:
                    print(e)

                self.device.node_map.AcquisitionStop.execute()

                # Flash the queue for image acquisition process.
                self._data_stream.flush_buffer_queue(
                    ACQ_QUEUE_TYPE_LIST.ACQ_QUEUE_ALL_DISCARD
                )

                # Unregister the registered event.
                self._event_manager.unregister_event()

                #
                self._release_data_stream()

            #
            self._initialize_buffers()

            for statistics in self._statistics_list:
                statistics.reset()

    def reset_statistics(self):
        self._current_width = self.device.node_map.Width.value
        self._current_height = self.device.node_map.Height.value
        self._current_pixel_format = self.device.node_map.PixelFormat.value

    def add_cti_file(self, file_path: str):
        """
        Adds a CTI file to work with to the CTI file list.

        :param file_path: Set a file path to the target CTI file.

        :return: None.
        """
        if file_path not in self._cti_files:
            self._cti_files.append(file_path)

    def remove_cti_file(self, file_path: str):
        """
        Remove the specified CTI file from the CTI file list.

        :param file_path: Set a file path to the target CTI file.

        :return: None.
        """
        if file_path in self._cti_files:
            self._cti_files.remove(file_path)

    def discard_cti_files(self):
        """
        Remove all CTI files in the CTI file list.

        :return: None.
        """

        self._cti_files = []

    def _open_gentl_producers(self):
        #
        for file_path in self._cti_files:
            producer = GenTLProducer.create_producer()
            try:
                producer.open(file_path)
            except ClosedException as e:
                print(e)
            else:
                self._producers.append(producer)

    def _open_systems(self):
        for producer in self._producers:
            system = producer.create_system()
            try:
                system.open()
            except ClosedException as e:
                print(e)
            else:
                self._systems.append(system)

    def reset(self):
        """
        Initializes the Harvester object.

        :return: None.
        """
        self.discard_cti_files()
        self._release_gentl_producers()

    def _release_gentl_producers(self):
        #
        self._release_systems()

        #
        for producer in self._producers:
            if producer and producer.is_open():
                producer.close()

        #
        self._producers = []

    def _release_systems(self):
        #
        self._release_interfaces()

        #
        for system in self._systems:
            if system is not None and system.is_open():
                system.close()

        #
        self._systems = []

    def _release_interfaces(self):
        #
        self._release_device_info_list()

        #
        if self._interfaces is not None:
            for iface in self._interfaces:
                if iface.is_open():
                    iface.close()

        #
        self._interfaces = []

    def _release_device_info_list(self):
        #
        self.disconnect_device()

        #
        if self.device_info_list is not None:
            self._device_info_list = []

    def _release_data_stream(self):
        #
        self._release_buffers()

        #
        if self._data_stream:
            if self._data_stream.is_open():
                self._data_stream.close()

        #
        self._data_stream = None
        self._event_manager = None

    def _release_buffers(self):
        #
        revoked_buffers = []
        if self._data_stream:
            for i, buffer in enumerate(self._announced_buffers):
                revoked_buffers.append(self._data_stream.revoke_buffer(buffer))

        self._announced_buffers = revoked_buffers

    def _initialize_buffers(self):
        self._fetched_buffers = []
        self._has_acquired_1st_image = False

    def update_device_info_list(self):
        """
        Updates the device information list. You'll have to call this method
        every time you added CTI files or plugged/unplugged devices.

        :return: None.
        """
        #
        self._release_gentl_producers()

        try:
            self._open_gentl_producers()
            self._open_systems()
            #
            for system in self._systems:
                #
                system.update_interface_info_list(self.timeout_for_update)

                #
                for i_info in system.interface_info_list:
                    iface = i_info.create_interface()
                    iface.open()
                    iface.update_device_info_list(self.timeout_for_update)
                    self._interfaces.append(iface)
                    for d_info in iface.device_info_list:
                        self.device_info_list.append(d_info)

        except LoadLibraryException as e:
            print(e)
            self._has_revised_device_list = False
        else:
            self._has_revised_device_list = True


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# ----------------------------------------------------------------------------
#
# Copyright 2018 EMVA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------


# Standard library imports

# Related third party imports

import numpy as np
from vispy import gloo
from vispy import app
from vispy.gloo import Program
from vispy.util.transforms import ortho

# Local application/library specific imports
from harvesters._private.core.helper.system import is_running_on_macos


class Canvas(app.Canvas):
    def __init__(
            self,
            harvester_core=None,
            width=640, height=480,
            fps=50.,
            background_color='gray',
            vertex_shader=None,
            fragment_shader=None
    ):
        """
        NOTE: fps should be smaller than or equal to 50 fps. If it's exceed
        VisPy drags down the acquisition performance. This is the issue we
        have to investigate.
        """

        self._vertex_shader = vertex_shader if vertex_shader else """
            // Uniforms
            uniform mat4 u_model;
            uniform mat4 u_view;
            uniform mat4 u_projection;

            // Attributes
            attribute vec2 a_position;
            attribute vec2 a_texcoord;

            // Varyings
            varying vec2 v_texcoord;

            // Main
            void main (void)
            {
                v_texcoord = a_texcoord;
                gl_Position = u_projection * u_view * u_model * vec4(a_position, 0.0, 1.0);
            }
        """

        self._fragment_shader = fragment_shader if fragment_shader else """
            varying vec2 v_texcoord;
            uniform sampler2D texture;
            void main()
            {
                gl_FragColor = texture2D(texture, v_texcoord);
            }
        """

        #
        self._harvester_core = harvester_core

        #
        app.Canvas.__init__(
            self, size=(width, height), vsync=True, autoswap=True
        )

        #
        self._program = None
        self._data = None
        self._coordinate = None
        self._origin = None
        self._translate = 0.
        self._latest_translate = self._translate
        self._magnification = 1.

        #
        self._background_color = background_color
        self._has_filled_texture = False
        self._width, self._height = width, height

        #
        self._is_dragging = False

        # If it's True , the canvas keeps image acquisition but do not
        # draw images on the canvas.
        self._pause_drawing = False

        # Apply shaders.
        self.set_shaders(
            vertex_shader=self._vertex_shader,
            fragment_shader=self._fragment_shader
        )

        #
        self._timer = app.Timer(1./fps, connect=self.update, start=True)

    def set_shaders(self, vertex_shader=None, fragment_shader=None):
        #
        vs = vertex_shader if vertex_shader else self._vertex_shader
        fs = fragment_shader if fragment_shader else self._fragment_shader
        self._program = Program(vs, fs, count=4)

        #
        self._data = np.zeros(
            4, dtype=[
                ('a_position', np.float32, 2),
                ('a_texcoord', np.float32, 2)
            ]
        )

        #
        self._data['a_texcoord'] = np.array(
            [[0., 1.], [1., 1.], [0., 0.], [1., 0.]]
        )

        #
        self._program['u_model'] = np.eye(4, dtype=np.float32)
        self._program['u_view'] = np.eye(4, dtype=np.float32)

        #
        self._coordinate = [0, 0]
        self._origin = [0, 0]

        #
        self._program['texture'] = np.zeros(
            (self._height, self._width), dtype='uint8'
        )

        #
        self.apply_magnification()

    def set_rect(self, width, height):
        #
        self._has_filled_texture = False

        #
        updated = False

        #
        if self._width != width or self._height != height:
            self._width = width
            self._height = height
            updated = True

        #
        if updated:
            self.apply_magnification()

    def on_draw(self, event):
        # Clear the canvas in gray.
        gloo.clear(color=self._background_color)
        self._update_texture()

    def _update_texture(self):
        # Fetch a buffer.
        try:
            with self._harvester_core.fetch_buffer(timeout_ms=0.1) as buffer:
                # Set the image as the texture of our canvas.
                if not self._pause_drawing and buffer:
                    # Update the canvas size if needed.
                    self.set_rect(buffer.image.width, buffer.image.height)
                    self._program['texture'] = buffer.image.ndarray

                # Draw the texture.
                self._draw()

        except AttributeError:
            # Harvester Core has not started image acquisition so
            # calling fetch_buffer() raises AttributeError because
            # None object is used for the with statement.

            # Update on June 15th, 2018:
            # According to a VisPy developer, they have not finished
            # porting VisPy to PyQt5. Once they finished the development
            # we should try it out if it gives us the maximum refresh rate.
            # See the following URL to check the latest information:
            #
            #     https://github.com/vispy/vispy/issues/1394

            # Draw the texture.
            self._draw()

    def _draw(self):
        self._program.draw('triangle_strip')

    def on_resize(self, event):
        self.apply_magnification()

    def apply_magnification(self):
        #
        canvas_w, canvas_h = self.physical_size
        gloo.set_viewport(0, 0, canvas_w, canvas_h)

        #
        ratio = self._magnification
        w, h = self._width, self._height

        self._program['u_projection'] = ortho(
            self._coordinate[0],
            canvas_w * ratio + self._coordinate[0],
            self._coordinate[1],
            canvas_h * ratio + self._coordinate[1],
            -1, 1
        )

        x, y = int((canvas_w * ratio - w) / 2), int((canvas_h * ratio - h) / 2)  # centering x & y

        #
        self._data['a_position'] = np.array(
            [[x, y], [x + w, y], [x, y + h], [x + w, y + h]]
        )

        #
        self._program.bind(gloo.VertexBuffer(self._data))

    def on_mouse_wheel(self, event):
        self._translate += event.delta[1]
        power = 7. if is_running_on_macos() else 5.  # 2 ** power
        stride = 4. if is_running_on_macos() else 7.
        translate = self._translate
        translate = min(power * stride, translate)
        translate = max(-power * stride, translate)
        self._translate = translate
        self._magnification = 2 ** -(self._translate / stride)
        if self._latest_translate != self._translate:
            self.apply_magnification()
            self._latest_translate = self._translate

    def on_mouse_press(self, event):
        self._is_dragging = True
        self._origin = event.pos

    def on_mouse_release(self, event):
        self._is_dragging = False

    def on_mouse_move(self, event):
        if self._is_dragging:
            adjustment = 2. if is_running_on_macos() else 1.
            ratio = self._magnification * adjustment
            delta = event.pos - self._origin
            self._origin = event.pos
            self._coordinate[0] -= (delta[0] * ratio)
            self._coordinate[1] += (delta[1] * ratio)
            self.apply_magnification()

    def pause_drawing(self, pause=True):
        self._pause_drawing = pause

    def toggle_drawing(self):
        self._pause_drawing = False if self._pause_drawing else True

    @property
    def is_pausing(self):
        return True if self._pause_drawing else False

    def resume_drawing(self):
        self._pause_drawing = False

    @property
    def background_color(self):
        return self._background_color

    @background_color.setter
    def background_color(self, color):
        self._background_color = color




#!/usr/bin/env python3
# ----------------------------------------------------------------------------
#
# Copyright 2018 EMVA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------


# Standard library imports
import time

# Related third party imports

# Local application/library specific imports


class Statistics:
    def __init__(self):
        #
        super().__init__()

        #
        self._time_base = 0
        self._time_elapsed = 0
        self._timestamp_base = 0
        self._has_acquired_1st_timestamp = False
        self._fps = 0.
        self._num_images = 0
        self._fps_max = 0.

    def set_timestamp(self, timestamp, frequency):
        if not self._has_acquired_1st_timestamp:
            self._timestamp_base = timestamp
            self._has_acquired_1st_timestamp = True
        else:
            diff = timestamp - self._timestamp_base
            if diff > 0:
                fps = (self._num_images - 1) * frequency / diff
                if fps > self._fps_max:
                    self._fps_max = fps
                self._fps = fps
            else:
                self._fps = 0.

    def reset(self):
        self._time_base = 0
        self._timestamp_base = 0
        self._has_acquired_1st_timestamp = False
        self._fps = 0.
        self._num_images = 0
        self._fps_max = 0.

    def increment_num_images(self, num=1):
        if self._has_acquired_1st_timestamp:
            self._time_elapsed = time.time() - self._time_base
        else:
            self._time_base = time.time()
        self._num_images += num

    @property
    def fps(self):
        return self._fps

    @property
    def fps_max(self):
        return self._fps_max

    @property
    def num_images(self):
        return self._num_images

    @property
    def elapsed_time_s(self):
        return self._time_elapsed




#!/usr/bin/env python3
# ----------------------------------------------------------------------------
#
# Copyright 2018 EMVA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------


# Standard library imports
from threading import Thread

# Related third party imports

# Local application/library specific imports
from harvesters._private.core.thread_ import ThreadBase


class PyThread(ThreadBase):
    def __init__(self, mutex=None, worker=None):
        #
        super().__init__(mutex=mutex)

        #
        self._thread = None
        self._worker = worker

    def _start(self):
        # Create a Thread object. The object is not reusable.
        self._thread = _PyThreadImpl(
            base=self,
            worker=self._worker
        )

        # Start running its worker method.
        self._thread.start()

    def stop(self):
        #
        if self._thread is None:
            return

        # Prepare to terminate the worker method.
        self._thread.stop()

        # Wait until the run methods is terminated.
        self._thread.join()

    def acquire(self):
        #
        if self._thread is None:
            return None

        #
        return self._thread.acquire()

    def release(self):
        #
        if self._thread is None:
            return

        #
        self._thread.release()

    @property
    def worker(self):
        #
        if self._thread is None:
            return None

        #
        return self._thread.worker

    @worker.setter
    def worker(self, obj):
        #
        if self._thread is None:
            return

        #
        self._thread.worker = obj

    @property
    def mutex(self):
        return self._mutex


class _PyThreadImpl(Thread):
    def __init__(self, base=None, worker=None):
        #
        super().__init__()

        #
        self._worker = worker
        self._base = base

    def stop(self):
        with self._base.mutex:
            self._base.is_running = False

    def run(self):
        """
        Runs its worker method.

        This method will be terminated once its parent's is_running
        property turns False.
        """
        while self._base.is_running:
            if self._worker:
                self._worker()

    def acquire(self):
        return self._base.mutex.acquire()

    def release(self):
        self._base.mutex.release()

    @property
    def worker(self):
        return self._worker

    @worker.setter
    def worker(self, obj):
        self._worker = obj



#!/usr/bin/env python3
# ----------------------------------------------------------------------------
#
# Copyright 2018 EMVA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------


# Standard library imports

# Related third party imports
import numpy as np

# Local application/library specific imports
from harvesters._private.core.image import Image


class Buffer:
    def __init__(self, data_stream=None, gentl_buffer=None, node_map=None, image: np.ndarray=None):
        #
        super().__init__()

        #
        self._data_stream = data_stream
        self._gentl_buffer = gentl_buffer
        self._node_map = node_map
        self._image = Image(self, image)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Queue the buffer when it goes outside of the scope.
        if self._data_stream and self._gentl_buffer:
            self._data_stream.queue_buffer(self._gentl_buffer)

    @property
    def image(self) -> Image:
        return self._image

    @property
    def data_stream(self):
        return self._data_stream

    @property
    def gentl_buffer(self):
        return self._gentl_buffer

    @property
    def node_map(self):
        return self._node_map





#!/usr/bin/env python3
# ----------------------------------------------------------------------------
#
# Copyright 2018 EMVA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------


# Standard library imports
import sys

# Related third party imports
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QMainWindow, QApplication, QTreeView, \
    QAction, QComboBox, QLineEdit, QLabel, QShortcut

from genicam2.genapi import EVisibility

# Local application/library specific imports
from harvesters._private.frontend.helper import compose_tooltip
from harvesters._private.frontend.pyqt5.action import Action
from harvesters._private.frontend.pyqt5.feature_tree import \
    FeatureEditDelegate, FilterProxyModel, FeatureTreeModel
from harvesters._private.frontend.pyqt5.helper import get_system_font
from harvesters._private.frontend.pyqt5.icon import Icon

"""

If you got into a trouble relate to model, the following tool could give
you a hint. ModelTest is a Python script and it tests your model and report
the result:

https://github.com/bgr/PyQt5_modeltest

"""


class ActionExpandAll(Action):
    def __init__(self, icon, title, parent=None):
        #
        super().__init__(icon, title, parent=parent)

    def _execute(self):
        self.parent().expand_all()

    def update(self):
        pass


class ActionCollapseAll(Action):
    def __init__(self, icon, title, parent=None):
        #
        super().__init__(icon, title, parent=parent)

    def _execute(self):
        self.parent().collapse_all()

    def update(self):
        pass


class AttributeController(QMainWindow):
    _visibility_dict = {
        'Beginner': EVisibility.Beginner,
        'Expert': EVisibility.Expert,
        'Guru': EVisibility.Guru,
        'All': EVisibility.Invisible,
    }

    def __init__(self, node_map, parent=None):
        #
        super().__init__(parent=parent)

        #
        self.setWindowTitle('Attribute Controller')

        #
        self._view = QTreeView()
        self._view.setFont(get_system_font())

        #
        self._node_map = node_map
        self._model = FeatureTreeModel(
            node_map=self._node_map,
            thread=self.parent().harvester_core.thread_image_acquisition,
        )

        #
        self._proxy = FilterProxyModel()
        self._proxy.setSourceModel(self._model)
        self._proxy.setDynamicSortFilter(False)

        #
        self._delegate = FeatureEditDelegate(proxy=self._proxy)
        self._view.setModel(self._proxy)
        self._view.setItemDelegate(self._delegate)

        #
        unit = 260
        for i in range(2):
            self._view.setColumnWidth(i, unit)

        w, h = 700, 600
        self._view.setGeometry(100, 100, w, h)

        self.setCentralWidget(self._view)
        self.setGeometry(100, 100, unit * 2, 640)

        self._combo_box_visibility = None
        self._line_edit_search_box = None

        #
        self._setup_toolbars()

    def _setup_toolbars(self):
        #
        group_filter = self.addToolBar('Node Visibility')
        group_manipulation = self.addToolBar('Node Tree Manipulation')

        #
        label_visibility = QLabel()
        label_visibility.setText('Visibility')
        label_visibility.setFont(get_system_font())

        #
        self._combo_box_visibility = QComboBox()
        self._combo_box_visibility.setSizeAdjustPolicy(
            QComboBox.AdjustToContents
        )

        items = ('Beginner', 'Expert', 'Guru', 'All')
        for item in items:
            self._combo_box_visibility.addItem(item)

        shortcut_key = 'Ctrl+v'
        shortcut = QShortcut(QKeySequence(shortcut_key), self)

        def show_popup():
            self._combo_box_visibility.showPopup()

        shortcut.activated.connect(show_popup)

        self._combo_box_visibility.setToolTip(
            compose_tooltip('Filter the nodes to show', shortcut_key)
        )
        self._combo_box_visibility.setFont(get_system_font())
        self._combo_box_visibility.currentIndexChanged.connect(
            self._invalidate_feature_tree_by_visibility
        )

        #
        button_expand_all = ActionExpandAll(
            Icon('expand.png'), 'Expand All', self
        )
        shortcut_key = 'Ctrl+e'
        button_expand_all.setToolTip(
            compose_tooltip('Expand the node tree', shortcut_key)
        )
        button_expand_all.setShortcut(shortcut_key)
        button_expand_all.toggle()

        #
        button_collapse_all = ActionCollapseAll(
            Icon('collapse.png'), 'Collapse All', self
        )
        shortcut_key = 'Ctrl+c'
        button_collapse_all.setToolTip(
            compose_tooltip('Collapse the node tree', shortcut_key)
        )
        button_collapse_all.setShortcut(shortcut_key)
        button_collapse_all.toggle()

        #
        label_search = QLabel()
        label_search.setText('RegEx Search')
        label_search.setFont(get_system_font())

        #
        self._line_edit_search_box = QLineEdit()
        self._line_edit_search_box.setFont(get_system_font())
        self._line_edit_search_box.textEdited.connect(
            self._invalidate_feature_tree_by_keyword
        )

        #
        group_filter.addWidget(label_visibility)
        group_filter.addWidget(self._combo_box_visibility)
        group_filter.addWidget(label_search)
        group_filter.addWidget(self._line_edit_search_box)
        group_filter.setStyleSheet('QToolBar{spacing:6px;}')

        #
        group_manipulation.addAction(button_expand_all)
        group_manipulation.addAction(button_collapse_all)

        #
        group_manipulation.actionTriggered[QAction].connect(
            self.on_button_clicked_action
        )

    def _invalidate_feature_tree_by_visibility(self):
        visibility = self._visibility_dict[
            self._combo_box_visibility.currentText()
        ]
        self._proxy.setVisibility(visibility)
        self._view.expandAll()

    @pyqtSlot('QString')
    def _invalidate_feature_tree_by_keyword(self, keyword):
        self._proxy.setKeyword(keyword)
        self._view.expandAll()

    @staticmethod
    def on_button_clicked_action(action):
        action.execute()

    def expand_all(self):
        self._view.expandAll()

    def collapse_all(self):
        self._view.collapseAll()

    def resize_column_width(self):
        for i in range(self._model.columnCount()):
            self._view.resizeColumnToContents(i)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    about = AttributeController()
    about.show()
    sys.exit(app.exec_())
